"""
Utility functions for GCM mode.
Implements Galois field arithmetic for GCM authentication.

@see BouncyCastle GCMUtil.java
"""

from ....util.pack import Pack


class GCMUtil:
    """Utility functions for GCM mode authentication."""
    
    BLOCK_SIZE = 16
    E1 = 0xe1000000
    
    @staticmethod
    def xor(block: bytearray, val: bytes) -> None:
        """XOR two blocks in-place."""
        for i in range(16):
            block[i] ^= val[i]
    
    @staticmethod
    def multiply(x_bytes: bytes, y_bytes: bytes) -> bytes:
        """
        Galois field multiplication in GF(2^128).
        Multiplies two 128-bit blocks.
        
        This implements the algorithm from BouncyCastle's GCMUtil.java multiply method.
        Using 32-bit int processing.
        """
        # Convert to 32-bit int arrays (4 ints = 128 bits)
        x = [0] * 4
        y = [0] * 4
        
        for i in range(4):
            offset = i * 4
            x[i] = ((x_bytes[offset] << 24) | (x_bytes[offset + 1] << 16) | 
                   (x_bytes[offset + 2] << 8) | x_bytes[offset + 3])
            y[i] = ((y_bytes[offset] << 24) | (y_bytes[offset + 1] << 16) | 
                   (y_bytes[offset + 2] << 8) | y_bytes[offset + 3])
            
            # Convert to signed 32-bit
            if x[i] & 0x80000000:
                x[i] -= 0x100000000
            if y[i] & 0x80000000:
                y[i] -= 0x100000000
        
        y0, y1, y2, y3 = y[0], y[1], y[2], y[3]
        z0 = z1 = z2 = z3 = 0
        
        # Process each bit of x
        for i in range(4):
            bits = x[i]
            for j in range(32):
                # Arithmetic shift: -1 if MSB set, 0 otherwise
                m1 = bits >> 31
                bits = (bits << 1) & 0xffffffff
                if bits & 0x80000000:
                    bits -= 0x100000000
                
                z0 ^= (y0 & m1)
                z1 ^= (y1 & m1)
                z2 ^= (y2 & m1)
                z3 ^= (y3 & m1)
                
                # Shift y right with reduction polynomial
                m2 = ((y3 << 31) >> 8) & 0xffffffff
                if m2 & 0x80000000:
                    m2 -= 0x100000000
                
                y3 = ((y3 & 0xffffffff) >> 1) | ((y2 << 31) & 0xffffffff)
                y2 = ((y2 & 0xffffffff) >> 1) | ((y1 << 31) & 0xffffffff)
                y1 = ((y1 & 0xffffffff) >> 1) | ((y0 << 31) & 0xffffffff)
                y0 = (((y0 & 0xffffffff) >> 1) ^ (m2 & GCMUtil.E1)) & 0xffffffff
                
                # Convert back to signed if needed
                if y0 & 0x80000000:
                    y0 -= 0x100000000
                if y1 & 0x80000000:
                    y1 -= 0x100000000
                if y2 & 0x80000000:
                    y2 -= 0x100000000
                if y3 & 0x80000000:
                    y3 -= 0x100000000
        
        # Convert result back to bytes
        result = bytearray(16)
        vals = [z0, z1, z2, z3]
        for i in range(4):
            val = vals[i] & 0xffffffff
            offset = i * 4
            result[offset] = (val >> 24) & 0xff
            result[offset + 1] = (val >> 16) & 0xff
            result[offset + 2] = (val >> 8) & 0xff
            result[offset + 3] = val & 0xff
        
        return bytes(result)
    
    @staticmethod
    def increment(counter: bytearray) -> None:
        """Increment the rightmost 32 bits of a counter block."""
        c = 1
        for i in range(15, 11, -1):
            c += counter[i] & 0xff
            counter[i] = c & 0xff
            c >>= 8



"""Padding schemes for block ciphers."""

from .pkcs7_padding import PKCS7Padding
from .zero_byte_padding import ZeroBytePadding
from .iso10126_padding import ISO10126Padding
from .iso7816_4_padding import ISO7816d4Padding
from .padded_buffered_block_cipher import PaddedBufferedBlockCipher

__all__ = ['PKCS7Padding', 'ZeroBytePadding', 'ISO10126Padding', 'ISO7816d4Padding', 'PaddedBufferedBlockCipher']



"""
Cipher Block Chaining (CBC) mode.

Reference: org.bouncycastle.crypto.modes.CBCBlockCipher
           src/crypto/modes/CBCBlockCipher.ts (sm-js-bc)
"""

from typing import Union, List
from sm_bc.crypto.cipher_parameters import CipherParameters
from sm_bc.crypto.params.parameters_with_iv import ParametersWithIV
from sm_bc.exceptions import DataLengthException


class CBCBlockCipher:
    """
    Cipher Block Chaining (CBC) mode implementation.
    
    CBC mode chains blocks together by XORing each plaintext block with the
    previous ciphertext block before encryption. The first block is XORed with
    the initialization vector (IV).
    
    Reference: NIST SP 800-38A
    """
    
    def __init__(self, cipher):
        """
        Initialize CBC mode with a block cipher.
        
        Args:
            cipher: The underlying block cipher (must implement BlockCipher interface)
        """
        self.cipher = cipher
        self.block_size = cipher.get_block_size()
        
        self.IV = bytearray(self.block_size)
        self.cbcV = bytearray(self.block_size)
        self.cbcNextV = bytearray(self.block_size)
        
        self.encrypting = False
    
    def get_underlying_cipher(self):
        """
        Return the underlying block cipher.
        
        Returns:
            The underlying block cipher
        """
        return self.cipher
    
    def init(self, for_encryption: bool, params: CipherParameters) -> None:
        """
        Initialize the cipher and possibly the IV.
        
        If an IV isn't passed as part of the parameter, the IV will be all zeros.
        
        Args:
            for_encryption: True for encryption, False for decryption
            params: The key and other data required by the cipher
            
        Raises:
            ValueError: If IV length doesn't match block size
        """
        old_encrypting = self.encrypting
        self.encrypting = for_encryption
        
        if isinstance(params, ParametersWithIV):
            iv = params.get_iv()
            
            if len(iv) != self.block_size:
                raise ValueError('initialization vector must be the same length as block size')
            
            self.IV[:] = iv
            params = params.get_parameters()
        else:
            # No IV provided, use all zeros
            for i in range(len(self.IV)):
                self.IV[i] = 0
        
        self.reset()
        
        # If params is None, it's an IV change only (key is to be reused)
        if params is not None:
            self.cipher.init(for_encryption, params)
        elif old_encrypting != for_encryption:
            raise ValueError('cannot change encrypting state without providing key')
    
    def get_algorithm_name(self) -> str:
        """
        Return the algorithm name and mode.
        
        Returns:
            The name of the underlying algorithm followed by "/CBC"
        """
        return self.cipher.get_algorithm_name() + '/CBC'
    
    def get_block_size(self) -> int:
        """
        Return the block size of the underlying cipher.
        
        Returns:
            The block size in bytes
        """
        return self.cipher.get_block_size()
    
    def process_block(self, input_data: Union[bytes, bytearray], in_off: int,
                     output: Union[bytearray, List[int]], out_off: int) -> int:
        """
        Process one block of input.
        
        Args:
            input_data: The array containing the input data
            in_off: Offset into the input array where data starts
            output: The array the output data will be copied into
            out_off: The offset into the output array where output will start
            
        Returns:
            The number of bytes processed and produced
            
        Raises:
            DataLengthException: If buffer too short
        """
        if self.encrypting:
            return self._encrypt_block(input_data, in_off, output, out_off)
        else:
            return self._decrypt_block(input_data, in_off, output, out_off)
    
    def reset(self) -> None:
        """Reset the chaining vector back to the IV and reset the underlying cipher."""
        self.cbcV[:] = self.IV
        for i in range(len(self.cbcNextV)):
            self.cbcNextV[i] = 0
        self.cipher.reset()
    
    def _encrypt_block(self, input_data: Union[bytes, bytearray], in_off: int,
                      output: Union[bytearray, List[int]], out_off: int) -> int:
        """
        Do the appropriate chaining step for CBC mode encryption.
        
        Args:
            input_data: The array containing the data to be encrypted
            in_off: Offset into the input array where data starts
            output: The array the encrypted data will be copied into
            out_off: The offset into the output array where output will start
            
        Returns:
            The number of bytes processed and produced
            
        Raises:
            DataLengthException: If input buffer too short
        """
        if in_off + self.block_size > len(input_data):
            raise DataLengthException('input buffer too short')
        
        # XOR the cbcV and the input, then encrypt the cbcV
        for i in range(self.block_size):
            self.cbcV[i] ^= input_data[in_off + i]
        
        length = self.cipher.process_block(self.cbcV, 0, output, out_off)
        
        # Copy ciphertext to cbcV
        self.cbcV[:] = output[out_off:out_off + self.block_size]
        
        return length
    
    def _decrypt_block(self, input_data: Union[bytes, bytearray], in_off: int,
                      output: Union[bytearray, List[int]], out_off: int) -> int:
        """
        Do the appropriate chaining step for CBC mode decryption.
        
        Args:
            input_data: The array containing the data to be decrypted
            in_off: Offset into the input array where data starts
            output: The array the decrypted data will be copied into
            out_off: The offset into the output array where output will start
            
        Returns:
            The number of bytes processed and produced
            
        Raises:
            DataLengthException: If input buffer too short
        """
        if in_off + self.block_size > len(input_data):
            raise DataLengthException('input buffer too short')
        
        # Save the ciphertext block for next round
        self.cbcNextV[:] = input_data[in_off:in_off + self.block_size]
        
        length = self.cipher.process_block(input_data, in_off, output, out_off)
        
        # XOR the cbcV and the output
        for i in range(self.block_size):
            output[out_off + i] ^= self.cbcV[i]
        
        # Swap the back up buffer into next position
        tmp = self.cbcV
        self.cbcV = self.cbcNextV
        self.cbcNextV = tmp
        
        return length



"""
Elliptic curve algorithms utility class.

Provides common elliptic curve operations and algorithms.
"""

from .ec_point import ECPoint
from .ec_curve import ECCurve


class ECAlgorithms:
    """Utility class containing various elliptic curve algorithms."""
    
    @staticmethod
    def clean_point(curve: ECCurve, point: ECPoint) -> ECPoint:
        """
        Clean a point by validating it's on the curve and normalizing if needed.
        
        Args:
            curve: The elliptic curve
            point: The point to clean
            
        Returns:
            The cleaned (validated and normalized) point
            
        Raises:
            ValueError: If point is not on the expected curve or is invalid
        """
        if not curve.equals(point.curve):
            raise ValueError('Point is not on the expected curve')
        
        if not point.is_valid():
            raise ValueError('Point is not valid')
        
        return point.normalize()
    
    @staticmethod
    def sum_of_two_multiplies(P1: ECPoint, k1: int, P2: ECPoint, k2: int) -> ECPoint:
        """
        Calculate k1*P1 + k2*P2 efficiently.
        
        This method computes the sum of two scalar multiplications more efficiently
        than computing them separately and then adding.
        
        Args:
            P1: The first point
            k1: The first scalar
            P2: The second point  
            k2: The second scalar
            
        Returns:
            The result point k1*P1 + k2*P2
            
        Raises:
            ValueError: If points are not on the same curve
        """
        if not P1.curve.equals(P2.curve):
            raise ValueError('Points must be on the same curve')
        
        # Use Shamir's trick for efficient dual scalar multiplication
        # This is a simplified implementation - could be optimized further
        result1 = P1.multiply(k1)
        result2 = P2.multiply(k2)
        
        return result1.add(result2)
    
    @staticmethod
    def is_point_at_infinity(point: ECPoint) -> bool:
        """
        Check if a point is the point at infinity.
        
        Args:
            point: The point to check
            
        Returns:
            True if the point is at infinity
        """
        return point.is_infinity()
    
    @staticmethod
    def are_on_same_curve(P1: ECPoint, P2: ECPoint) -> bool:
        """
        Validate that two points are on the same curve.
        
        Args:
            P1: First point
            P2: Second point
            
        Returns:
            True if both points are on the same curve
        """
        return P1.curve.equals(P2.curve)



"""
ISO 10126 Padding implementation.

ISO 10126 padding fills with random bytes, except the last byte
which contains the padding length.

Reference: Withdrawn ISO/IEC 10116:2006 (now deprecated but still in use)
"""

from typing import Union
import secrets


class ISO10126Padding:
    """
    ISO 10126 Padding scheme.
    
    Pads data with random bytes, with the last byte indicating padding length.
    This provides some security through randomness but is now deprecated in
    favor of PKCS#7.
    
    Format:
    - Fill with random bytes
    - Last byte contains padding length (1-255)
    
    Example for 3-byte padding: [random, random, 0x03]
    
    Note: This standard was withdrawn in 2007 but is still used in some systems.
    """
    
    def __init__(self):
        """Initialize ISO 10126 Padding."""
        pass
    
    def add_padding(self, data: Union[bytes, bytearray], block_size: int) -> bytearray:
        """
        Add ISO 10126 padding to data.
        
        Args:
            data: The data to pad
            block_size: The block size in bytes
            
        Returns:
            Padded data
            
        Raises:
            ValueError: If block_size is invalid
        """
        if block_size < 1 or block_size > 255:
            raise ValueError(f'Invalid block size: {block_size}')
        
        # Calculate padding length
        padding_length = block_size - (len(data) % block_size)
        
        # Create padded data
        padded = bytearray(data)
        
        # Add random bytes (except last byte)
        if padding_length > 1:
            padded.extend(secrets.token_bytes(padding_length - 1))
        
        # Last byte is padding length
        padded.append(padding_length)
        
        return padded
    
    def remove_padding(self, data: Union[bytes, bytearray], block_size: int) -> bytearray:
        """
        Remove ISO 10126 padding from data.
        
        Args:
            data: The padded data
            block_size: The block size in bytes
            
        Returns:
            Unpadded data
            
        Raises:
            ValueError: If padding is invalid or block_size is invalid
        """
        if block_size < 1 or block_size > 255:
            raise ValueError(f'Invalid block size: {block_size}')
        
        if len(data) == 0:
            raise ValueError('Cannot remove padding from empty data')
        
        if len(data) % block_size != 0:
            raise ValueError('Data length is not a multiple of block size')
        
        # Get padding length from last byte
        padding_length = data[-1]
        
        # Validate padding length
        if padding_length < 1 or padding_length > block_size:
            raise ValueError(f'Invalid padding length: {padding_length}')
        
        if len(data) < padding_length:
            raise ValueError('Padding length exceeds data length')
        
        # Remove padding
        return bytearray(data[:-padding_length])
    
    def get_padded_length(self, data_length: int, block_size: int) -> int:
        """
        Calculate the length after padding.
        
        Args:
            data_length: The original data length
            block_size: The block size in bytes
            
        Returns:
            Length after padding
        """
        padding_length = block_size - (data_length % block_size)
        return data_length + padding_length



from .ec_key_parameters import ECKeyParameters
from .ec_domain_parameters import ECDomainParameters
from ...math.ec_point import ECPoint

class ECPublicKeyParameters(ECKeyParameters):
    def __init__(self, q: ECPoint, parameters: ECDomainParameters):
        super().__init__(False, parameters)
        self.q = self.validate(q)

    def validate(self, q: ECPoint) -> ECPoint:
        # BC checks if q is valid on curve
        if not q.is_valid():
             raise ValueError("Point not on curve")
        return q
    
    def get_Q(self) -> ECPoint:
        """Get the public key point Q."""
        return self.q



from typing import Protocol
from sm_bc.math.ec_point import ECPoint

class ECMultiplier(Protocol):
    def multiply(self, p: ECPoint, k: int) -> ECPoint:
        ...

class AbstractECMultiplier(ECMultiplier):
    def multiply(self, p: ECPoint, k: int) -> ECPoint:
        if p.is_infinity or k == 0:
            return p.curve.get_infinity()
        
        if k < 0:
            p = p.negate()
            k = -k
            
        return self.multiply_positive(p, k)

    def multiply_positive(self, p: ECPoint, k: int) -> ECPoint:
        raise NotImplementedError
        
class SimpleMultiplier(AbstractECMultiplier):
    def multiply_positive(self, p: ECPoint, k: int) -> ECPoint:
        return p.multiply(k)


class FixedPointCombMultiplier(AbstractECMultiplier):
    """
    Fixed-point comb multiplier for efficient scalar multiplication.
    This is an optimized multiplier that uses precomputation for fixed points.
    For now, it uses the standard multiplication algorithm.
    """
    
    def multiply_positive(self, p: ECPoint, k: int) -> ECPoint:
        """
        Multiply point p by scalar k using comb method.
        
        Args:
            p: The point to multiply
            k: The scalar multiplier (positive)
            
        Returns:
            The result point k*p
        """
        # For now, use standard multiplication
        # TODO: Implement actual comb method with precomputation
        return p.multiply(k)



from typing import Union, List, Optional
from sm_bc.crypto.digest import Digest
from sm_bc.crypto.digests.sm3_digest import SM3Digest
from sm_bc.crypto.signers.dsa_encoding import DSAEncoding, StandardDSAEncoding
from sm_bc.crypto.signers.dsa_k_calculator import DSAKCalculator, RandomDSAKCalculator
from sm_bc.crypto.cipher_parameters import CipherParameters
from sm_bc.crypto.params.ec_key_parameters import ECKeyParameters
from sm_bc.crypto.params.ec_private_key_parameters import ECPrivateKeyParameters
from sm_bc.crypto.params.ec_public_key_parameters import ECPublicKeyParameters
from sm_bc.crypto.params.parameters_with_id import ParametersWithID
from sm_bc.crypto.params.parameters_with_random import ParametersWithRandom
from sm_bc.math.ec_multiplier import SimpleMultiplier, ECMultiplier
from sm_bc.math.ec_point import ECPoint
from sm_bc.util.secure_random import SecureRandom

class SM2Signer:
    def __init__(self, digest: Digest = None, encoding: DSAEncoding = None, k_calculator: DSAKCalculator = None):
        self.digest = digest if digest else SM3Digest()
        self.encoding = encoding if encoding else StandardDSAEncoding()
        self.k_calculator = k_calculator if k_calculator else RandomDSAKCalculator()
        
        self.ec_params = None
        self.pub_key_point: Optional[ECPoint] = None
        self.ec_key: Optional[ECKeyParameters] = None
        self.z: bytes = b''
    
    def get_algorithm_name(self) -> str:
        """Return the algorithm name."""
        return 'SM2'
    
    def init(self, for_signing: bool, param: CipherParameters) -> None:
        self.for_signing = for_signing
        
        base_param = param
        user_id = b"1234567812345678"
        
        if isinstance(base_param, ParametersWithID):
            user_id = base_param.id
            base_param = base_param.parameters
            
        if for_signing:
            if isinstance(base_param, ParametersWithRandom):
                r_param = base_param
                base_param = r_param.parameters
                self.k_calculator.init(base_param.parameters.n, r_param.random)
            else:
                self.k_calculator.init(base_param.parameters.n, SecureRandom())
                
            if not isinstance(base_param, ECPrivateKeyParameters):
                raise ValueError("ECPrivateKeyParameters required for signing")
            self.ec_key = base_param
        else:
             if not isinstance(base_param, ECPublicKeyParameters):
                raise ValueError("ECPublicKeyParameters required for verification")
             self.ec_key = base_param
             
        self.ec_params = self.ec_key.parameters
        
        # Calculate Z
        # We need public key point for Z. 
        # If signing, we derive it from d * G ? No, usually we need public key for Z calculation even when signing?
        # The standard says Z uses User's Public Key.
        # So if we only have private key, we must calculate public key point: P = d * G.
        
        if for_signing:
            self.pub_key_point = self.ec_params.g.multiply(self.ec_key.d).normalize()
        else:
            self.pub_key_point = self.ec_key.q.normalize()
            
        self.z = self.get_z(user_id)
        
        self.digest.reset()
        self.digest.update_bytes(self.z, 0, len(self.z))

    def update(self, b: int) -> None:
        self.digest.update(b)

    def update_bytes(self, b: Union[bytes, bytearray, List[int]], off: int, len_: int) -> None:
        self.digest.update_bytes(b, off, len_)

    def generate_signature(self) -> bytes:
        if not self.for_signing:
            raise ValueError("Not initialized for signing")
            
        n = self.ec_params.n
        e_hash = bytearray(self.digest.get_digest_size())
        self.digest.do_final(e_hash, 0)
        
        e = int.from_bytes(e_hash, 'big')
        d = self.ec_key.d
        
        multiplier = self.create_base_point_multiplier()
        
        while True:
            k = self.k_calculator.next_k()
            
            # (x1, y1) = k * G
            p1 = multiplier.multiply(self.ec_params.g, k).normalize()
            x1 = p1.x.to_big_integer()
            
            r = (e + x1) % n
            if r == 0 or (r + k) == n:
                continue
                
            # s = (1 + d)^-1 * (k - r * d)
            d_plus_1 = (1 + d) % n
            d_plus_1_inv = pow(d_plus_1, -1, n)
            
            s = (d_plus_1_inv * (k - r * d)) % n
            
            if s == 0:
                continue
                
            return self.encoding.encode(n, r, s)

    def verify_signature(self, signature: bytes) -> bool:
        if self.for_signing:
            raise ValueError("Not initialized for verification")
            
        n = self.ec_params.n
        try:
            r, s = self.encoding.decode(n, signature)
        except Exception:
            return False
            
        if r < 1 or r >= n or s < 1 or s >= n:
            return False
            
        e_hash = bytearray(self.digest.get_digest_size())
        self.digest.do_final(e_hash, 0)
        e = int.from_bytes(e_hash, 'big')
        
        t = (r + s) % n
        if t == 0:
            return False
            
        # (x1, y1) = s*G + t*P
        # P is pub_key_point
        p1 = self.ec_params.g.multiply(s).add(self.pub_key_point.multiply(t)).normalize()
        
        if p1.is_infinity:
            return False
            
        x1 = p1.x.to_big_integer()
        expected_r = (e + x1) % n
        
        return expected_r == r

    def reset(self) -> None:
        self.digest.reset()
        if self.z:
            self.digest.update_bytes(self.z, 0, len(self.z))

    def get_z(self, user_id: bytes) -> bytes:
        digest = SM3Digest()
        
        # entl * 8 (bits)
        length = len(user_id) * 8
        digest.update(length >> 8 & 0xFF)
        digest.update(length & 0xFF)
        
        digest.update_bytes(user_id, 0, len(user_id))
        
        self.add_field_element(digest, self.ec_params.curve.a)
        self.add_field_element(digest, self.ec_params.curve.b)
        self.add_field_element(digest, self.ec_params.g.x)
        self.add_field_element(digest, self.ec_params.g.y)
        self.add_field_element(digest, self.pub_key_point.x)
        self.add_field_element(digest, self.pub_key_point.y)
        
        output = bytearray(digest.get_digest_size())
        digest.do_final(output, 0)
        return output

    def add_field_element(self, digest: Digest, v) -> None:
        val = v.to_big_integer()
        size = (self.ec_params.curve.get_field_size() + 7) // 8
        digest.update_bytes(val.to_bytes(size, 'big'), 0, size)
        
    def create_base_point_multiplier(self) -> ECMultiplier:
        return SimpleMultiplier()



from typing import Union, MutableSequence, List

class Pack:
    """
    Utility class for packing/unpacking bytes to integers.
    Note: Unlike Java/JS, this implementation returns Unsigned Integers by default
    to suit Python's handling of bitwise operations.
    """

    @staticmethod
    def big_endian_to_int(bs: Union[bytes, bytearray, List[int]], off: int) -> int:
        """
        Reads a 32-bit unsigned integer from the byte buffer at the specified offset.
        """
        n = bs[off] << 24
        n |= (bs[off + 1] & 0xff) << 16
        n |= (bs[off + 2] & 0xff) << 8
        n |= (bs[off + 3] & 0xff)
        return n

    @staticmethod
    def int_to_big_endian(val: int, bs: MutableSequence[int], off: int) -> None:
        """
        Writes a 32-bit integer to the byte buffer at the specified offset.
        """
        bs[off] = (val >> 24) & 0xff
        bs[off + 1] = (val >> 16) & 0xff
        bs[off + 2] = (val >> 8) & 0xff
        bs[off + 3] = val & 0xff

    @staticmethod
    def big_endian_to_long(bs: Union[bytes, bytearray, List[int]], off: int) -> int:
        """
        Reads a 64-bit unsigned integer from the byte buffer.
        """
        hi = Pack.big_endian_to_int(bs, off)
        lo = Pack.big_endian_to_int(bs, off + 4)
        return (hi << 32) | (lo & 0xffffffff)

    @staticmethod
    def long_to_big_endian(val: int, bs: MutableSequence[int], off: int) -> None:
        """
        Writes a 64-bit integer to the byte buffer.
        """
        # In Python, right shift of negative numbers behaves differently (arithmetic shift),
        # but for packing we typically deal with the bit pattern.
        # We mask to ensure we get the bits we expect.
        
        # High 32 bits
        hi = (val >> 32) & 0xffffffff
        # Low 32 bits
        lo = val & 0xffffffff
        
        Pack.int_to_big_endian(hi, bs, off)
        Pack.int_to_big_endian(lo, bs, off + 4)



"""
High-level cipher interface for easy encryption/decryption.

This module provides convenient wrapper classes that combine block ciphers,
cipher modes, and padding schemes into easy-to-use interfaces.
"""

from typing import Union, Optional
from .engines.sm4_engine import SM4Engine
from .modes import CBCBlockCipher, SICBlockCipher, OFBBlockCipher, CFBBlockCipher
from .paddings import PKCS7Padding, ISO7816d4Padding, ZeroBytePadding, ISO10126Padding
from .params.key_parameter import KeyParameter
from .params.parameters_with_iv import ParametersWithIV


class SM4Cipher:
    """
    High-level SM4 cipher interface.
    
    Provides convenient methods for encrypting and decrypting data with SM4
    using various cipher modes and padding schemes.
    
    Example:
        # CBC mode with PKCS#7 padding
        cipher = SM4Cipher(mode='CBC', padding='PKCS7')
        cipher.init(True, key, iv)
        ciphertext = cipher.encrypt(plaintext)
        
        cipher.init(False, key, iv)
        plaintext = cipher.decrypt(ciphertext)
    """
    
    MODES = {
        'ECB': ('direct', None),  # Direct SM4Engine
        'CBC': ('block', CBCBlockCipher),
        'CTR': ('stream', SICBlockCipher),
        'OFB': ('stream_bitblock', OFBBlockCipher),
        'CFB': ('stream_bitblock', CFBBlockCipher),
    }
    
    PADDINGS = {
        'PKCS7': PKCS7Padding,
        'ISO7816-4': ISO7816d4Padding,
        'ISO10126': ISO10126Padding,
        'ZERO': ZeroBytePadding,
        'NONE': None,
    }
    
    def __init__(self, mode: str = 'CBC', padding: str = 'PKCS7'):
        """
        Initialize SM4 cipher.
        
        Args:
            mode: Cipher mode ('ECB', 'CBC', 'CTR', 'OFB', 'CFB')
            padding: Padding scheme ('PKCS7', 'ISO7816-4', 'ISO10126', 'ZERO', 'NONE')
        
        Raises:
            ValueError: If mode or padding is invalid
        """
        if mode.upper() not in self.MODES:
            raise ValueError(f"Invalid mode: {mode}. Must be one of {list(self.MODES.keys())}")
        
        if padding.upper() not in self.PADDINGS:
            raise ValueError(f"Invalid padding: {padding}. Must be one of {list(self.PADDINGS.keys())}")
        
        self.mode_name = mode.upper()
        self.padding_name = padding.upper()
        
        # Create base engine
        base_engine = SM4Engine()
        
        # Get mode configuration
        mode_type, mode_class = self.MODES[self.mode_name]
        
        # Wrap with mode if needed
        if mode_type == 'direct':
            self.engine = base_engine
            self.is_stream = False
        elif mode_type == 'block':
            self.engine = mode_class(base_engine)
            self.is_stream = False
        elif mode_type == 'stream':
            self.engine = mode_class(base_engine)
            self.is_stream = True
        elif mode_type == 'stream_bitblock':
            # OFB and CFB need bit_block_size parameter (128 bits = 16 bytes)
            self.engine = mode_class(base_engine, 128)
            self.is_stream = True
        
        # Create padding if needed
        if self.PADDINGS[self.padding_name] is not None:
            self.padding = self.PADDINGS[self.padding_name]()
        else:
            self.padding = None
        
        self.block_size = 16  # SM4 block size
        self.initialized = False
        self.encrypting = False
    
    def init(self, encrypting: bool, key: Union[bytes, bytearray], 
             iv: Optional[Union[bytes, bytearray]] = None):
        """
        Initialize the cipher for encryption or decryption.
        
        Args:
            encrypting: True for encryption, False for decryption
            key: 16-byte encryption key
            iv: Initialization vector (required for CBC, OFB, CFB, CTR modes)
        
        Raises:
            ValueError: If key length is invalid or IV is missing when required
        """
        if len(key) != 16:
            raise ValueError(f"Invalid key length: {len(key)}. SM4 requires 16 bytes.")
        
        self.encrypting = encrypting
        
        # Check if IV is required
        if self.mode_name in ['CBC', 'CTR', 'OFB', 'CFB']:
            if iv is None:
                raise ValueError(f"IV is required for {self.mode_name} mode")
            if len(iv) != 16:
                raise ValueError(f"Invalid IV length: {len(iv)}. Must be 16 bytes.")
            params = ParametersWithIV(KeyParameter(key), iv)
        else:
            params = KeyParameter(key)
        
        self.engine.init(encrypting, params)
        self.initialized = True
    
    def encrypt(self, plaintext: Union[bytes, bytearray]) -> bytearray:
        """
        Encrypt plaintext.
        
        Args:
            plaintext: Data to encrypt
        
        Returns:
            Encrypted data (ciphertext)
        
        Raises:
            RuntimeError: If cipher is not initialized or initialized for decryption
        """
        if not self.initialized:
            raise RuntimeError("Cipher not initialized. Call init() first.")
        
        if not self.encrypting:
            raise RuntimeError("Cipher initialized for decryption, not encryption")
        
        # For stream ciphers, no padding needed
        if self.is_stream:
            data = bytearray(plaintext)
            output = bytearray(len(data))
            if hasattr(self.engine, 'process_bytes'):
                self.engine.process_bytes(data, 0, len(data), output, 0)
            else:
                # Fallback for stream ciphers without process_bytes
                for i in range(0, len(data), self.block_size):
                    self.engine.process_block(data, i, output, i)
            return output
        
        # For block ciphers, handle padding
        if self.padding is not None:
            data = self.padding.add_padding(plaintext, self.block_size)
        else:
            data = bytearray(plaintext)
            # Check block alignment
            if len(data) % self.block_size != 0:
                raise ValueError(f"Data length must be multiple of {self.block_size} when padding is disabled")
        
        # Encrypt
        output = bytearray(len(data))
        for i in range(0, len(data), self.block_size):
            self.engine.process_block(data, i, output, i)
        
        return output
    
    def decrypt(self, ciphertext: Union[bytes, bytearray]) -> bytearray:
        """
        Decrypt ciphertext.
        
        Args:
            ciphertext: Data to decrypt
        
        Returns:
            Decrypted data (plaintext)
        
        Raises:
            RuntimeError: If cipher is not initialized or initialized for encryption
            ValueError: If ciphertext length is invalid
        """
        if not self.initialized:
            raise RuntimeError("Cipher not initialized. Call init() first.")
        
        if self.encrypting:
            raise RuntimeError("Cipher initialized for encryption, not decryption")
        
        # For stream ciphers, no padding to remove
        if self.is_stream:
            output = bytearray(len(ciphertext))
            if hasattr(self.engine, 'process_bytes'):
                self.engine.process_bytes(ciphertext, 0, len(ciphertext), output, 0)
            else:
                # Fallback for stream ciphers without process_bytes
                for i in range(0, len(ciphertext), self.block_size):
                    self.engine.process_block(ciphertext, i, output, i)
            return output
        
        # For block ciphers, check alignment
        if len(ciphertext) % self.block_size != 0:
            raise ValueError(f"Ciphertext length must be multiple of {self.block_size}")
        
        # Decrypt
        output = bytearray(len(ciphertext))
        for i in range(0, len(ciphertext), self.block_size):
            self.engine.process_block(ciphertext, i, output, i)
        
        # Remove padding if needed
        if self.padding is not None and hasattr(self.padding, 'remove_padding'):
            output = self.padding.remove_padding(output, self.block_size)
        
        return output
    
    def reset(self):
        """Reset the cipher to initial state."""
        if hasattr(self.engine, 'reset'):
            self.engine.reset()
        self.initialized = False
    
    def get_algorithm_name(self) -> str:
        """Get the algorithm name."""
        return f"SM4/{self.mode_name}/{self.padding_name if self.padding_name != 'NONE' else 'NoPadding'}"


def create_sm4_cipher(mode: str = 'CBC', padding: str = 'PKCS7') -> SM4Cipher:
    """
    Factory function to create SM4 cipher.
    
    Args:
        mode: Cipher mode ('ECB', 'CBC', 'CTR', 'OFB', 'CFB')
        padding: Padding scheme ('PKCS7', 'ISO7816-4', 'ISO10126', 'ZERO', 'NONE')
    
    Returns:
        Configured SM4Cipher instance
    
    Example:
        cipher = create_sm4_cipher('CBC', 'PKCS7')
        cipher.init(True, key, iv)
        ciphertext = cipher.encrypt(plaintext)
    """
    return SM4Cipher(mode=mode, padding=padding)



"""
ECB (Electronic Codebook) 模式实现

警告：ECB 模式不安全，不应在生产环境中使用。
相同的明文块总是加密为相同的密文块，这会泄露信息模式。

此实现仅用于：
- 与旧系统的兼容性
- 测试和教学目的

参考: org.bouncycastle.crypto.modes.ECBBlockCipher
"""

from sm_bc.crypto.cipher_parameters import CipherParameters
from sm_bc.exceptions import DataLengthException


class ECBBlockCipher:
    """ECB模式块密码"""

    def __init__(self, cipher):
        """
        初始化ECB模式
        
        Args:
            cipher: 底层块密码
        """
        self.cipher = cipher
        self.block_size = cipher.get_block_size()

    def get_underlying_cipher(self):
        """返回底层密码"""
        return self.cipher

    def init(self, encrypting: bool, params: CipherParameters) -> None:
        """
        初始化密码
        
        Args:
            encrypting: True表示加密，False表示解密
            params: 密码参数
        """
        self.cipher.init(encrypting, params)

    def get_algorithm_name(self) -> str:
        """返回算法名称"""
        return self.cipher.get_algorithm_name() + '/ECB'

    def get_block_size(self) -> int:
        """返回块大小（字节）"""
        return self.block_size

    def process_block(
        self,
        input_bytes: bytes,
        in_off: int,
        output: bytearray,
        out_off: int
    ) -> int:
        """
        处理一个数据块
        
        Args:
            input_bytes: 输入数据
            in_off: 输入偏移量
            output: 输出缓冲区
            out_off: 输出偏移量
            
        Returns:
            处理的字节数
        """
        if in_off + self.block_size > len(input_bytes):
            raise DataLengthException('input buffer too short')

        if out_off + self.block_size > len(output):
            raise DataLengthException('output buffer too short')

        return self.cipher.process_block(input_bytes, in_off, output, out_off)

    def reset(self) -> None:
        """重置密码到初始状态"""
        self.cipher.reset()



from .ec_key_parameters import ECKeyParameters
from .ec_domain_parameters import ECDomainParameters

class ECPrivateKeyParameters(ECKeyParameters):
    def __init__(self, d: int, parameters: ECDomainParameters):
        super().__init__(True, parameters)
        self.d = d
    
    def get_d(self) -> int:
        """Get the private key scalar d."""
        return self.d



"""
带填充的缓冲块密码

此类包装一个BlockCipher并应用填充，使其能够处理任意长度的数据。

参考: org.bouncycastle.crypto.paddings.PaddedBufferedBlockCipher
"""

from typing import Any
from sm_bc.crypto.cipher_parameters import CipherParameters
from sm_bc.exceptions.data_length_exception import DataLengthException
from sm_bc.exceptions.invalid_cipher_text_exception import InvalidCipherTextException


class PaddedBufferedBlockCipher:
    """带填充的缓冲块密码"""
    
    def __init__(self, cipher: Any, padding: Any):
        """
        构造函数
        
        Args:
            cipher: 底层块密码
            padding: 填充方案
        """
        self.cipher = cipher
        self.padding = padding
        self.buf = bytearray(cipher.get_block_size())
        self.buf_off = 0
        self.for_encryption = False
    
    def get_underlying_cipher(self) -> Any:
        """获取底层密码"""
        return self.cipher
    
    def init(self, encrypting: bool, params: CipherParameters) -> None:
        """
        初始化密码
        
        Args:
            encrypting: True 表示加密，False 表示解密
            params: 密码参数
        """
        self.for_encryption = encrypting
        self.reset()
        self.cipher.init(encrypting, params)
    
    def get_block_size(self) -> int:
        """返回块大小"""
        return self.cipher.get_block_size()
    
    def get_output_size(self, length: int) -> int:
        """
        获取输出大小
        
        Args:
            length: 输入长度
            
        Returns:
            输出将需要的字节数
        """
        total = length + self.buf_off
        left_over = total % len(self.buf)
        
        if left_over == 0:
            if self.for_encryption:
                return total + len(self.buf)
            return total
        
        return total - left_over + len(self.buf)
    
    def get_update_output_size(self, length: int) -> int:
        """
        获取可更新输出大小
        
        Args:
            length: 输入长度
            
        Returns:
            更新时将产生的字节数
        """
        total = length + self.buf_off
        left_over = total % len(self.buf)
        return total - left_over
    
    def process_byte(self, input_byte: int, output: bytearray, out_off: int) -> int:
        """
        处理一个字节
        
        Args:
            input_byte: 输入字节
            output: 输出缓冲区
            out_off: 输出偏移量
            
        Returns:
            写入输出的字节数
        """
        result_len = 0
        
        self.buf[self.buf_off] = input_byte
        self.buf_off += 1
        
        if self.buf_off == len(self.buf):
            result_len = self.cipher.process_block(self.buf, 0, output, out_off)
            self.buf_off = 0
        
        return result_len
    
    def process_bytes(
        self,
        input_data: bytes,
        in_off: int,
        length: int,
        output: bytearray,
        out_off: int
    ) -> int:
        """
        处理字节数组
        
        Args:
            input_data: 输入数据
            in_off: 输入偏移量
            length: 要处理的字节数
            output: 输出缓冲区
            out_off: 输出偏移量
            
        Returns:
            写入输出的字节数
        """
        if length < 0:
            raise ValueError('Cannot have a negative input length')
        
        block_size = self.get_block_size()
        gap_len = len(self.buf) - self.buf_off
        
        if length > gap_len:
            self.buf[self.buf_off:self.buf_off + gap_len] = input_data[in_off:in_off + gap_len]
            
            result_len = self.cipher.process_block(self.buf, 0, output, out_off)
            
            self.buf_off = 0
            length -= gap_len
            in_off += gap_len
            
            while length > len(self.buf):
                result_len += self.cipher.process_block(input_data, in_off, output, out_off + result_len)
                
                length -= block_size
                in_off += block_size
            
            self.buf[self.buf_off:self.buf_off + length] = input_data[in_off:in_off + length]
            
            self.buf_off += length
            
            return result_len
        else:
            self.buf[self.buf_off:self.buf_off + length] = input_data[in_off:in_off + length]
            
            self.buf_off += length
            
            return 0
    
    def do_final(self, output: bytearray, out_off: int) -> int:
        """
        完成处理
        
        Args:
            output: 输出缓冲区
            out_off: 输出偏移量
            
        Returns:
            写入输出的字节数
        """
        block_size = self.cipher.get_block_size()
        result_len = 0
        
        if self.for_encryption:
            if self.buf_off == block_size:
                if out_off + 2 * block_size > len(output):
                    self.reset()
                    raise DataLengthException('output buffer too short')
                
                result_len = self.cipher.process_block(self.buf, 0, output, out_off)
                self.buf_off = 0
            
            self.padding.add_padding(self.buf, self.buf_off)
            
            result_len += self.cipher.process_block(self.buf, 0, output, out_off + result_len)
            
            self.reset()
        else:
            if self.buf_off == block_size:
                result_len = self.cipher.process_block(self.buf, 0, self.buf, 0)
                self.buf_off = 0
            else:
                self.reset()
                raise DataLengthException('last block incomplete in decryption')
            
            try:
                count = self.padding.pad_count(self.buf)
                
                result_len -= count
                
                output[out_off:out_off + result_len] = self.buf[0:result_len]
            except Exception as e:
                self.reset()
                if isinstance(e, (DataLengthException, InvalidCipherTextException)):
                    raise
                raise InvalidCipherTextException(str(e))
            finally:
                self.reset()
        
        return result_len
    
    def reset(self) -> None:
        """重置密码到初始状态"""
        for i in range(len(self.buf)):
            self.buf[i] = 0
        self.buf_off = 0
        
        self.cipher.reset()



"""
Utility functions for integer operations.

Provides bit manipulation and other integer utilities similar to Java's Integer class.
All operations treat integers as 32-bit signed values.

Reference: src/util/Integers.ts (sm-js-bc)
"""


class Integers:
    """
    Integer utility functions for bit manipulation.
    
    All methods treat integers as 32-bit signed values, matching Java/TypeScript behavior.
    """
    
    @staticmethod
    def _to_int32(i: int) -> int:
        """Convert Python int to 32-bit signed integer."""
        # Mask to 32 bits and interpret as signed
        i = i & 0xFFFFFFFF
        if i >= 0x80000000:
            i -= 0x100000000
        return i
    
    @staticmethod
    def number_of_leading_zeros(i: int) -> int:
        """
        Returns the number of zero bits preceding the highest-order one-bit.
        Returns 32 if the number is zero.
        
        Args:
            i: The value whose number of leading zero bits to compute
            
        Returns:
            The number of zero bits preceding the highest-order one-bit (0-32)
        """
        # Convert to 32-bit integer
        i = Integers._to_int32(i)
        
        if i == 0:
            return 32
        
        # Handle negative numbers (MSB is set)
        if i < 0:
            return 0
        
        n = 1
        if (i >> 16) == 0:
            n += 16
            i <<= 16
        if (i >> 24) == 0:
            n += 8
            i <<= 8
        if (i >> 28) == 0:
            n += 4
            i <<= 4
        if (i >> 30) == 0:
            n += 2
            i <<= 2
        n -= (i >> 31) & 1
        
        return n
    
    @staticmethod
    def bit_count(i: int) -> int:
        """
        Returns the number of one-bits in the two's complement binary representation.
        
        Args:
            i: The value whose bits to count
            
        Returns:
            The number of one-bits
        """
        # Convert to 32-bit integer
        i = Integers._to_int32(i)
        # Make unsigned for bit operations
        i = i & 0xFFFFFFFF
        
        # Brian Kernighan's algorithm
        i = i - ((i >> 1) & 0x55555555)
        i = (i & 0x33333333) + ((i >> 2) & 0x33333333)
        i = (i + (i >> 4)) & 0x0F0F0F0F
        i = i + (i >> 8)
        i = i + (i >> 16)
        return i & 0x3F
    
    @staticmethod
    def rotate_left(i: int, distance: int) -> int:
        """
        Returns the value obtained by rotating left by the specified number of bits.
        
        Args:
            i: The value whose bits to rotate left
            distance: The number of bit positions to rotate by
            
        Returns:
            The value obtained by rotating left
        """
        # Convert to 32-bit integer
        i = Integers._to_int32(i)
        i = i & 0xFFFFFFFF  # Make unsigned
        distance = distance & 31  # Only use lower 5 bits
        
        result = ((i << distance) | (i >> (32 - distance))) & 0xFFFFFFFF
        return Integers._to_int32(result)
    
    @staticmethod
    def rotate_right(i: int, distance: int) -> int:
        """
        Returns the value obtained by rotating right by the specified number of bits.
        
        Args:
            i: The value whose bits to rotate right
            distance: The number of bit positions to rotate by
            
        Returns:
            The value obtained by rotating right
        """
        # Convert to 32-bit integer
        i = Integers._to_int32(i)
        i = i & 0xFFFFFFFF  # Make unsigned
        distance = distance & 31  # Only use lower 5 bits
        
        result = ((i >> distance) | (i << (32 - distance))) & 0xFFFFFFFF
        return Integers._to_int32(result)
    
    @staticmethod
    def number_of_trailing_zeros(i: int) -> int:
        """
        Returns the number of zero bits following the lowest-order one-bit.
        Returns 32 if the number is zero.
        
        Args:
            i: The value whose number of trailing zero bits to compute
            
        Returns:
            The number of zero bits following the lowest-order one-bit (0-32)
        """
        # Convert to 32-bit integer
        i = Integers._to_int32(i)
        
        if i == 0:
            return 32
        
        # Count trailing zeros using bit manipulation
        # Isolate the rightmost 1-bit: i & -i
        # Count position of that bit
        count = 0
        i_unsigned = i & 0xFFFFFFFF
        
        if (i_unsigned & 0xFFFF) == 0:
            count += 16
            i_unsigned >>= 16
        if (i_unsigned & 0xFF) == 0:
            count += 8
            i_unsigned >>= 8
        if (i_unsigned & 0xF) == 0:
            count += 4
            i_unsigned >>= 4
        if (i_unsigned & 0x3) == 0:
            count += 2
            i_unsigned >>= 2
        if (i_unsigned & 0x1) == 0:
            count += 1
        
        return count
    
    @staticmethod
    def highest_one_bit(i: int) -> int:
        """
        Returns the highest one bit of the specified number.
        
        Args:
            i: The value whose highest one bit to compute
            
        Returns:
            The highest one bit, or 0 if i is 0
        """
        # Convert to 32-bit integer
        i = Integers._to_int32(i)
        i = i & 0xFFFFFFFF  # Make unsigned
        
        i |= (i >> 1)
        i |= (i >> 2)
        i |= (i >> 4)
        i |= (i >> 8)
        i |= (i >> 16)
        result = (i - (i >> 1)) & 0xFFFFFFFF
        return Integers._to_int32(result)
    
    @staticmethod
    def lowest_one_bit(i: int) -> int:
        """
        Returns the lowest one bit of the specified number.
        
        Args:
            i: The value whose lowest one bit to compute
            
        Returns:
            The lowest one bit, or 0 if i is 0
        """
        # Convert to 32-bit integer
        i = Integers._to_int32(i)
        
        # Two's complement trick: i & -i isolates lowest bit
        result = (i & -i) & 0xFFFFFFFF
        return Integers._to_int32(result)



"""
PKCS#7 padding implementation.

Reference: org.bouncycastle.crypto.paddings.PKCS7Padding
"""

from typing import Union


class PKCS7Padding:
    """
    PKCS#7 padding - pads with the number of padding bytes.
    
    Reference: org.bouncycastle.crypto.paddings.PKCS7Padding
    """
    
    def __init__(self):
        """Initialize PKCS7 Padding."""
        pass
    
    def add_padding(self, data: Union[bytes, bytearray], block_size: int) -> bytearray:
        """
        Add PKCS7 padding to data.
        
        Args:
            data: The data to pad
            block_size: The block size in bytes
            
        Returns:
            Padded data
            
        Raises:
            ValueError: If block_size is invalid
        """
        if block_size < 1 or block_size > 255:
            raise ValueError(f'Invalid block size: {block_size}')
        
        # Calculate padding length
        padding_length = block_size - (len(data) % block_size)
        
        # Create padded data
        padded = bytearray(data)
        
        # Add padding bytes (all set to padding_length)
        padded.extend(bytes([padding_length] * padding_length))
        
        return padded
    
    def remove_padding(self, data: Union[bytes, bytearray], block_size: int) -> bytearray:
        """
        Remove PKCS7 padding from data.
        
        Args:
            data: The padded data
            block_size: The block size in bytes
            
        Returns:
            Unpadded data
            
        Raises:
            ValueError: If padding is invalid or block_size is invalid
        """
        if block_size < 1 or block_size > 255:
            raise ValueError(f'Invalid block size: {block_size}')
        
        if len(data) == 0:
            raise ValueError('Cannot remove padding from empty data')
        
        if len(data) % block_size != 0:
            raise ValueError('Data length is not a multiple of block size')
        
        # Get padding length from last byte
        padding_length = data[-1]
        
        # Validate padding length
        if padding_length < 1 or padding_length > block_size:
            raise ValueError(f'Invalid padding length: {padding_length}')
        
        if len(data) < padding_length:
            raise ValueError('Padding length exceeds data length')
        
        # Verify all padding bytes are correct
        for i in range(1, padding_length + 1):
            if data[-i] != padding_length:
                raise ValueError('Invalid padding bytes')
        
        # Remove padding
        return bytearray(data[:-padding_length])
    
    def get_padded_length(self, data_length: int, block_size: int) -> int:
        """
        Calculate the length after padding.
        
        Args:
            data_length: The original data length
            block_size: The block size in bytes
            
        Returns:
            Length after padding
        """
        padding_length = block_size - (data_length % block_size)
        return data_length + padding_length



"""
SM2 curve and domain parameters.

Based on GM/T 0003-2012 and GMNamedCurves.java
"""

from sm_bc.math.ec_curve import Fp as ECCurveFp
from sm_bc.math.ec_point import ECPoint
from sm_bc.crypto.params.ec_domain_parameters import ECDomainParameters
from sm_bc.util.secure_random import SecureRandom
from sm_bc.crypto.engines.sm2_engine import SM2Engine
from sm_bc.crypto.signers.sm2_signer import SM2Signer
from sm_bc.crypto.params.ec_public_key_parameters import ECPublicKeyParameters
from sm_bc.crypto.params.ec_private_key_parameters import ECPrivateKeyParameters
from sm_bc.crypto.params.parameters_with_random import ParametersWithRandom


class SM2:
    """SM2 elliptic curve parameters (SM2P256V1)."""
    
    # Prime p
    p = 0xFFFFFFFEFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF00000000FFFFFFFFFFFFFFFF
    
    # Curve coefficient a
    a = 0xFFFFFFFEFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF00000000FFFFFFFFFFFFFFFC
    
    # Curve coefficient b
    b = 0x28E9FA9E9D9F5E344D5A9E4BCF6509A7F39789F515AB8F92DDBCBD414D940E93
    
    # Base point order n
    n = 0xFFFFFFFEFFFFFFFFFFFFFFFFFFFFFFFF7203DF6B21C6052B53BBF40939D54123
    
    # Cofactor h
    h = 1
    
    _domain_params = None
    
    @classmethod
    def get_parameters(cls) -> ECDomainParameters:
        """Get SM2 domain parameters."""
        if cls._domain_params is None:
            # Create curve
            curve = ECCurveFp(cls.p, cls.a, cls.b, cls.n, cls.h)
            
            # Create base point G
            G = curve.create_point(cls.Gx, cls.Gy)
            
            # Create domain parameters
            cls._domain_params = ECDomainParameters(curve, G, cls.n, cls.h, None)
        
        return cls._domain_params
    
    @classmethod
    def get_curve(cls) -> ECCurveFp:
        """Get curve."""
        return cls.get_parameters().curve
    
    @classmethod
    def get_G(cls) -> ECPoint:
        """Get base point G."""
        return cls.get_parameters().g
    
    @classmethod
    def get_n(cls) -> int:
        """Get order n."""
        return cls.n
    
    @classmethod
    def get_h(cls) -> int:
        """Get cofactor h."""
        return cls.h
    
    # Base point G coordinates
    Gx = 0x32C4AE2C1F1981195F9904466A39C9948FE30BBFF2660BE1715A4589334C74C7
    Gy = 0xBC3736A2F4F6779C59BDCEE36B692153D0A9877CC62A474002DF32E52139F0A0
    
    @classmethod
    def validate_public_key(cls, Q: ECPoint) -> bool:
        """Validate public key point."""
        if Q.is_infinity:
            return False
        
        if not Q.is_valid():
            return False
        
        # Check [n]Q = O
        nQ = Q.multiply(cls.n)
        if not nQ.is_infinity:
            return False
        
        return True
    
    @classmethod
    def validate_private_key(cls, d: int) -> bool:
        """Validate private key."""
        return 0 < d < cls.n
    
    @classmethod
    def generate_key_pair(cls) -> dict:
        """
        Generate SM2 key pair.
        
        Returns:
            Dictionary containing private key (int) and public key coordinates {x, y}
        """
        random = SecureRandom()
        
        # Generate random private key d where 1 <= d < n
        while True:
            seed = random.generate_seed(32)  # 256 bits / 8 = 32 bytes
            d = int.from_bytes(seed, 'big')
            if 0 < d < cls.n:
                break
        
        # Calculate public key Q = [d]G
        G = cls.get_G()
        Q = G.multiply(d)
        
        if Q.is_infinity or not cls.validate_public_key(Q):
            raise ValueError('Generated invalid public key')
        
        affine_Q = Q.normalize()
        
        return {
            'private_key': d,
            'public_key': {
                'x': affine_Q.x.to_big_integer(),
                'y': affine_Q.y.to_big_integer()
            }
        }
    
    @classmethod
    def encrypt(cls, message: bytes | str, public_key: dict | int, public_key_y: int = None) -> bytes:
        """
        Encrypt plaintext using SM2 public key.
        
        Args:
            message: Message to encrypt (string or bytes)
            public_key: Public key object with x and y coordinates, or separate x coordinate
            public_key_y: Public key Y coordinate (if first parameter is not a dict)
        
        Returns:
            Encrypted ciphertext as bytes
        """
        # Convert string to bytes if needed
        if isinstance(message, str):
            message_bytes = message.encode('utf-8')
        else:
            message_bytes = message
        
        # Handle different parameter formats
        if isinstance(public_key, dict):
            public_key_x = public_key['x']
            public_key_y_value = public_key['y']
        elif isinstance(public_key, int) and isinstance(public_key_y, int):
            public_key_x = public_key
            public_key_y_value = public_key_y
        else:
            raise ValueError('Invalid public key format. Expected dict with x,y properties or separate x,y int values')
        
        # Create public key parameters
        curve = cls.get_curve()
        Q = curve.create_point(public_key_x, public_key_y_value)
        domain_params = cls.get_parameters()
        pub_key = ECPublicKeyParameters(Q, domain_params)
        
        # Create engine and initialize for encryption
        engine = SM2Engine()
        engine.init(True, ParametersWithRandom(pub_key, SecureRandom()))
        
        # Encrypt
        return engine.process_block(message_bytes, 0, len(message_bytes))
    
    @classmethod
    def decrypt(cls, ciphertext: bytes, private_key: int) -> bytes:
        """
        Decrypt ciphertext using SM2 private key.
        
        Args:
            ciphertext: Ciphertext to decrypt
            private_key: Private key
        
        Returns:
            Decrypted plaintext as bytes
        """
        # Create private key parameters
        domain_params = cls.get_parameters()
        priv_key = ECPrivateKeyParameters(private_key, domain_params)
        
        # Create engine and initialize for decryption
        engine = SM2Engine()
        engine.init(False, priv_key)
        
        # Decrypt
        return engine.process_block(ciphertext, 0, len(ciphertext))
    
    @classmethod
    def sign(cls, message: bytes | str, private_key: int) -> bytes:
        """
        Sign a message using SM2 private key.
        
        Args:
            message: Message to sign (string or bytes)
            private_key: Private key for signing
        
        Returns:
            Signature as bytes
        """
        # Convert string to bytes if needed
        if isinstance(message, str):
            message_bytes = message.encode('utf-8')
        else:
            message_bytes = message
        
        # Create private key parameters
        domain_params = cls.get_parameters()
        priv_key = ECPrivateKeyParameters(private_key, domain_params)
        
        # Create signer and initialize for signing
        signer = SM2Signer()
        signer.init(True, ParametersWithRandom(priv_key, SecureRandom()))
        
        # Update signer with message data
        signer.update_bytes(message_bytes, 0, len(message_bytes))
        
        # Generate signature
        return signer.generate_signature()
    
    @classmethod
    def verify(cls, message: bytes | str, signature: bytes, 
               public_key: dict | int, public_key_y: int = None) -> bool:
        """
        Verify a signature using SM2 public key.
        
        Args:
            message: Original message (string or bytes)
            signature: Signature to verify
            public_key: Public key object with x and y coordinates, or separate x coordinate
            public_key_y: Public key Y coordinate (if first parameter is not a dict)
        
        Returns:
            True if signature is valid, False otherwise
        """
        # Convert string to bytes if needed
        if isinstance(message, str):
            message_bytes = message.encode('utf-8')
        else:
            message_bytes = message
        
        # Handle different parameter formats
        if isinstance(public_key, dict):
            public_key_x = public_key['x']
            public_key_y_value = public_key['y']
        elif isinstance(public_key, int) and isinstance(public_key_y, int):
            public_key_x = public_key
            public_key_y_value = public_key_y
        else:
            raise ValueError('Invalid public key format. Expected dict with x,y properties or separate x,y int values')
        
        # Create public key parameters
        curve = cls.get_curve()
        Q = curve.create_point(public_key_x, public_key_y_value)
        domain_params = cls.get_parameters()
        pub_key = ECPublicKeyParameters(Q, domain_params)
        
        # Create signer and initialize for verification
        signer = SM2Signer()
        signer.init(False, pub_key)
        
        # Update signer with message data
        signer.update_bytes(message_bytes, 0, len(message_bytes))
        
        # Verify the signature
        return signer.verify_signature(signature)



"""
Utility functions for GCM mode.
Implements Galois field arithmetic for GCM authentication.
"""
from sm_bc.util.pack import Pack


class GCMUtil:
    """Utility functions for GCM mode"""
    
    BLOCK_SIZE = 16
    E1 = 0xe1000000
    
    @staticmethod
    def xor(block: bytearray, val: bytes) -> None:
        """XOR two blocks in-place"""
        for i in range(16):
            block[i] ^= val[i]
    
    @staticmethod
    def as_longs(x: bytes) -> tuple[int, int]:
        """Convert bytes to two 64-bit integers (big-endian)"""
        x0 = Pack.big_endian_to_long(x, 0)
        x1 = Pack.big_endian_to_long(x, 8)
        return (x0, x1)
    
    @staticmethod
    def from_longs(x0: int, x1: int) -> bytes:
        """Convert two 64-bit integers to bytes (big-endian)"""
        result = bytearray(16)
        Pack.long_to_big_endian(x0, result, 0)
        Pack.long_to_big_endian(x1, result, 8)
        return bytes(result)
    
    @staticmethod
    def multiply(x_bytes: bytes, y_bytes: bytes) -> bytes:
        """
        Galois field multiplication in GF(2^128).
        Multiplies two 128-bit blocks.
        """
        # Convert to 32-bit int arrays (4 ints = 128 bits)
        x = [0] * 4
        y = [0] * 4
        
        for i in range(4):
            offset = i * 4
            x[i] = (x_bytes[offset] << 24) | (x_bytes[offset + 1] << 16) | \
                   (x_bytes[offset + 2] << 8) | x_bytes[offset + 3]
            y[i] = (y_bytes[offset] << 24) | (y_bytes[offset + 1] << 16) | \
                   (y_bytes[offset + 2] << 8) | y_bytes[offset + 3]
        
        y0, y1, y2, y3 = y[0], y[1], y[2], y[3]
        z0, z1, z2, z3 = 0, 0, 0, 0
        
        # Process each bit of x
        for i in range(4):
            bits = x[i]
            for j in range(32):
                # Arithmetic shift: -1 if MSB set, 0 otherwise
                m1 = -1 if (bits & 0x80000000) else 0
                bits = (bits << 1) & 0xffffffff
                z0 ^= (y0 & m1) & 0xffffffff
                z1 ^= (y1 & m1) & 0xffffffff
                z2 ^= (y2 & m1) & 0xffffffff
                z3 ^= (y3 & m1) & 0xffffffff
                
                # Shift y right with reduction polynomial
                m2 = ((y3 << 31) >> 8) & 0xffffffff
                y3 = ((y3 >> 1) | (y2 << 31)) & 0xffffffff
                y2 = ((y2 >> 1) | (y1 << 31)) & 0xffffffff
                y1 = ((y1 >> 1) | (y0 << 31)) & 0xffffffff
                y0 = ((y0 >> 1) ^ (m2 & GCMUtil.E1)) & 0xffffffff
        
        # Convert result back to bytes
        result = bytearray(16)
        for i in range(4):
            val = [z0, z1, z2, z3][i]
            offset = i * 4
            result[offset] = (val >> 24) & 0xff
            result[offset + 1] = (val >> 16) & 0xff
            result[offset + 2] = (val >> 8) & 0xff
            result[offset + 3] = val & 0xff
        
        return bytes(result)
    
    @staticmethod
    def increment(counter: bytearray) -> None:
        """Increment the rightmost 32 bits of a counter block"""
        c = 1
        for i in range(15, 11, -1):
            c += counter[i] & 0xff
            counter[i] = c & 0xff
            c >>= 8
